@@ -124,13 +124,15 @@ class DebugDrawSettings:
 
 @dataclass
 class FrontendBaseSettings:
-    canvas_shape: tuple = (1200, 1200)
-
-    fps: int = 60
-    fixed_delta_t: bool = True  # If True, use a fixed delta time for the physics simulation
+    canvas_shape: tuple = (800, 800)
 
+    hertz: int = 60  # Physics update frequency
+    speed: float = 1.0  # Speed multiplier for the simulation
     substeps: int = 5
+
+    # depcrecated as as we center examples automatically
     ppm: float = 40.0  # Pixels per meter
+
     debug_draw: DebugDrawSettings = field(default_factory=DebugDrawSettings)
     multi_click_delay_ms: int = 350  # Delay in milliseconds to wait for multi-clicks
     headless: bool = False  # If True, run in headless mode (no GUI)
@@ -194,6 +196,25 @@ class TripleClickEvent(MouseEvent):
         super().__init__(world_position, handled)
 
 
+class KeyDownEvent(Event):
+    def __init__(self, key, ctrl=False, shift=False, meta=False, alt=False, handled=False):
+        super().__init__(handled)
+        self.key = key
+        self.ctrl = ctrl
+        self.shift = shift
+        self.meta = meta
+        self.alt = alt
+
+    def __repr__(self):
+        return f"KeyDownEvent(key={self.key}, ctrl={self.ctrl}, shift={self.shift}, meta={self.meta}, alt={self.alt})"
+
+
+class KeyUpEvent(Event):
+    def __init__(self, key, handled=False):
+        super().__init__(handled)
+        self.key = key
+
+
 class MultiClickHandler:
     def __init__(self, delayed_time_ms, on_click, on_double_click=None, on_triple_click=None):
         self.delayed_time = delayed_time_ms / 1000.0
@@ -286,6 +307,24 @@ class FrontendBase(ABC):
         self.sample_update_time = None
         self._multi_click_handler = None
 
+        # the last time when the world was updated
+        self.last_world_update_time = None
+
+        # the dt for the physics update
+        self.physics_update_dt = 1 / self.settings.hertz
+
+        self._is_paused = False
+
+        # store which key is currently pressed
+        self._pressed_keys = set()
+
+    def set_paused(self):
+        self._is_paused = True
+        self.last_world_update_time = None
+
+    def set_running(self):
+        self._is_paused = False
+
     def set_sample(self, sample_class, sample_settings=None):
         self.sample_class = sample_class
         self.sample_settings = sample_settings
@@ -340,65 +379,57 @@ class FrontendBase(ABC):
         # this is used to signal that the frontend is ready to run the sample
         self._set_new_sample(self.sample_class, self.sample_settings)
 
-    def _debug_draw_calls(self):
-        if self.settings.debug_draw.enabled:
-            self.debug_draw.begin_draw()
-
-        self.sample.pre_debug_draw()
-        t0 = time.time()
-        if self.settings.debug_draw.enabled:
-            self.sample.world.draw(self.debug_draw, call_begin_end=False)
-        self.acc_debug_draw_time += time.time() - t0
-
-        self.sample.post_debug_draw()
-
-        if self.settings.debug_draw.enabled:
-            self.debug_draw.end_draw()
-
-    def _sample_update_calls(self, dt):
-        # Update the world with the given time step
+    def update_physics_single_step(self):
+        dt = self.settings.speed / self.settings.hertz
         self.sample.pre_update(dt)
-        t0 = time.time()
         self.sample.update(dt)
-        self.sample_update_time = time.time() - t0
         self.sample.post_update(dt)
-        self.acc_update_time += self.sample_update_time
 
-    def update_and_draw(self, dt, single_step_mode=False):
-        if self.sample is None:
-            # if the sample is not set, we cannot update or draw it
+    def update_physics(self):
+        if self._is_paused:
+            # if we are paused, we don't update the physics
             return
+        expected_dt = 1 / self.settings.hertz
+        now = time.perf_counter()
+        if self.last_world_update_time is None:
+            self.update_physics_single_step()
+            self.last_world_update_time = now
+            return
+
+        dt = now - self.last_world_update_time
+
+        while dt >= expected_dt:
+            self.update_physics_single_step()
+            dt -= expected_dt
+            self.last_world_update_time += expected_dt
+
+    def update_frontend_logic(self):
+        if self._is_paused:
+            self._last_world_update_time = None
         # do we need to change the sample class?
         if self.change_sample_class_requested:
             self.change_sample_class_requested = False
             self.sample.post_run()
             self._set_new_sample(self.sample_class, self.sample_settings)
 
-        if self.is_paused():
-            # we still need to update the samples we want to do a single step
-            if single_step_mode:
-                self._sample_update_calls(dt)
-            self._debug_draw_calls()
-        else:
-            # click handler update
-            if self._multi_click_handler:
-                self._multi_click_handler.update()
-            self._sample_update_calls(dt)
-            self._debug_draw_calls()
-            self.iteration += 1
+        # click handler update
+        if self._multi_click_handler is not None:
+            self._multi_click_handler.update()
 
-    def on_play(self):
-        pass
+    def draw_physics(self):
+        t0 = time.time()
+        if self.settings.debug_draw.enabled:
+            self.debug_draw.begin_draw()
 
-    def on_pause(self):
-        pass
+        self.sample.pre_debug_draw()
+        if self.settings.debug_draw.enabled:
+            self.sample.world.draw(self.debug_draw, call_begin_end=False)
 
-    def single_step(self):
-        fps = self.settings.fps
-        if fps == 0:
-            fps = 60  # default to 60 FPS if not set
-        dt = 1 / fps
-        self.update_and_draw(dt, single_step_mode=True)
+        self.sample.post_debug_draw()
+
+        if self.settings.debug_draw.enabled:
+            self.debug_draw.end_draw()
+        self.acc_debug_draw_time += time.time() - t0
 
     def stop(self):
         self.set_sample(self.sample_class, self.sample_settings)
@@ -474,7 +505,7 @@ class FrontendBase(ABC):
         transform.offset = world_delta
 
     def is_paused(self):
-        return False
+        return self._is_paused
 
     @abstractmethod
     def drag_camera(self, delta):
@@ -491,3 +522,28 @@ class FrontendBase(ABC):
 
     def add_widget(self, element):
         pass
+
+    def is_key_pressed(self, key):
+        """Check if a key is pressed."""
+        return key in self._pressed_keys
+
+    def pressed_keys(self):
+        """Get the currently pressed keys set"""
+        return self._pressed_keys
+
+    def _on_key_down(self, event):
+        """Handle key down events."""
+        if event.key in self._pressed_keys:
+            # if the key is already pressed, we don't add it again
+            # this is because some frontend keep fireing the key down event
+            # when the key is pressed
+            # e.g. the ipycanvas frontend does this
+            return
+        self._pressed_keys.add(event.key)
+        self.sample.on_key_down(event)
+
+    def _on_key_up(self, event):
+        """Handle key up events."""
+        if event.key in self._pressed_keys:
+            self._pressed_keys.remove(event.key)
+        self.sample.on_key_up(event)
