@@ -23,6 +23,8 @@ from pyb2d3_sandbox.frontend_base import (
     MouseWheelEvent,
     MouseEnterEvent,
     MouseLeaveEvent,
+    KeyDownEvent,
+    KeyUpEvent,
 )
 import pyb2d3_sandbox.widgets as widgets
 
@@ -71,6 +73,36 @@ ICON_STOP = icons_fontawesome_6.ICON_FA_STOP
 ICON_FORWARD_STEP = icons_fontawesome_6.ICON_FA_FORWARD_STEP
 
 
+KEY_MAP = {
+    imgui.Key.space: "space",
+    imgui.Key.left_arrow: "left",
+    imgui.Key.right_arrow: "right",
+    imgui.Key.up_arrow: "up",
+    imgui.Key.down_arrow: "down",
+    imgui.Key.escape: "escape",
+    imgui.Key.enter: "enter",
+    imgui.Key.tab: "tab",
+    # ctrl keys
+    imgui.Key.left_ctrl: "ctrl",
+    imgui.Key.right_ctrl: "ctrl",
+    # shift keys
+    imgui.Key.left_shift: "shift",
+    imgui.Key.right_shift: "shift",
+    # meta keys
+    imgui.Key.left_super: "meta",
+    imgui.Key.right_super: "meta",
+    # alt keys
+    imgui.Key.left_alt: "alt",
+    imgui.Key.right_alt: "alt",
+    # Add letter keys
+    **{getattr(imgui.Key, f"{chr(i)}"): chr(i) for i in range(ord("a"), ord("z") + 1)},
+    # Add number keys
+    **{getattr(imgui.Key, f"_{i}"): str(i) for i in range(10)},
+}
+
+KEY_LIST = [(key, value) for key, value in KEY_MAP.items()]
+
+
 class OpenglFrontend(FrontendBase):
     @dataclass
     class Settings(FrontendBase.Settings):
@@ -87,10 +119,8 @@ class OpenglFrontend(FrontendBase):
         self.init_app()
         self._last_mouse_pos = None
         self._last_world_mouse_pos = None
-        self._center_when_ready = False
         self._was_inside_last_frame = False
 
-        self._is_paused = False
         self._just_a_single_frame = False
 
         self.debug_draw_option_names = [
@@ -112,8 +142,8 @@ class OpenglFrontend(FrontendBase):
 
         self._end_of_last_frame_time = time.time()
 
-    def is_paused(self):
-        return self._is_paused
+        # just some ui state
+        self.speed_ui_val = 0.0
 
     @property
     def weak_self(self):
@@ -134,7 +164,7 @@ class OpenglFrontend(FrontendBase):
         runner_params.callbacks.post_init = lambda: weak.post_gl_init()
 
         # Menu setup
-        runner_params.imgui_window_params.show_menu_bar = True
+        runner_params.imgui_window_params.show_menu_bar = False
         runner_params.callbacks.show_menus = lambda: weak.show_menus()
         runner_params.imgui_window_params.show_menu_app = False
 
@@ -149,7 +179,7 @@ class OpenglFrontend(FrontendBase):
         runner_params.docking_params = self.create_layout()
         runner_params.docking_params.main_dock_space_node_flags = imgui.DockNodeFlags_.none
 
-        # runner_params.callbacks.pre_new_frame = lambda: weak.on_pre_new_frame()
+        runner_params.callbacks.pre_new_frame = lambda: weak.on_pre_new_frame()
 
         # # on exit
         # runner_params.callbacks.on_exit = lambda: self.sample.destroy()
@@ -170,49 +200,29 @@ class OpenglFrontend(FrontendBase):
         #     #     setattr(state.show_dd, key, newvalue)
         #     imgui.end_menu()
 
-    def _late_center(self, size):
-        if self._center_when_ready:
-            # Center the sample when the debug draw is ready
-            self.debug_draw.camera.set_view(
-                self.debug_draw.camera.center,
-                self.debug_draw.camera.zoom,
-                size.x,
-                size.y,
-            )
-            self.center_sample(margin_px=self._margin_px)
-            self._center_when_ready = False
-            self._margin_px = None
-
     def handle_events(self, pos, io):
         if self._is_paused:
-            # print("OpenglFrontend.handle_events(): paused")
             return
+
+        # mouse events
         if imgui.is_window_hovered():
             if not self._was_inside_last_frame:
                 self.sample.on_mouse_enter(MouseEnterEvent())
             self._was_inside_last_frame = True
 
-            # print("OpenglFrontend.once_per_frame(): window hovered")
-
             mouse_pos = Vec2(io.mouse_pos.x, io.mouse_pos.y) - Vec2(pos.x, pos.y)
             world_pos = self.debug_draw.camera.convert_screen_to_world(mouse_pos)
 
-            # Only handle scroll when mouse is over simulation window
-            mouse_scroll = io.mouse_wheel
-            if mouse_scroll != 0.0:
+            if io.mouse_wheel != 0.0:
                 self.sample.on_mouse_wheel(
-                    MouseWheelEvent(delta=-mouse_scroll, world_position=world_pos)
+                    MouseWheelEvent(delta=-io.mouse_wheel, world_position=world_pos)
                 )
-
-            # print(" OpenglFrontend.once_per_frame(): mouse_pos:", mouse_pos, "world_pos:", world_pos)
-            # Check left mouse button events.
             if io.mouse_clicked[0]:
                 event = MouseDownEvent(world_position=world_pos)
                 self.sample.on_mouse_down(event)
                 self._multi_click_handler.handle_click(world_position=world_pos)
             elif io.mouse_down[0]:
                 delta = Vec2(io.mouse_delta.x, io.mouse_delta.y)
-
                 if delta[0] != 0 or delta[1] != 0:
                     event = MouseMoveEvent(
                         world_position=world_pos,
@@ -227,14 +237,34 @@ class OpenglFrontend(FrontendBase):
                 self.sample.on_mouse_leave(MouseLeaveEvent())
             self._was_inside_last_frame = False
 
+        # keyboard events
+        # get all keys which are **just pressed down in this frame**
+        # Check currently pressed keys
+        ikp = imgui.is_key_pressed
+        ikr = imgui.is_key_released
+        for key_code, key_name in KEY_LIST:
+            if ikp(key_code, repeat=False):
+                # check if ctrl, shift, meta, or alt are pressed
+                ctrl = imgui.get_io().key_ctrl
+                shift = imgui.get_io().key_shift
+                meta = imgui.get_io().key_super
+                alt = imgui.get_io().key_alt
+                self._on_key_down(
+                    KeyDownEvent(
+                        key=key_name,
+                        ctrl=ctrl,
+                        shift=shift,
+                        meta=meta,
+                        alt=alt,
+                    )
+                )
+            if ikr(key_code):
+                self._on_key_up(KeyUpEvent(key=key_name))
+
     def once_per_frame(self):
         if self.debug_draw is None or self.sample is None:
             return
 
-        # if self._is_paused:
-        #     # print("OpenglFrontend.once_per_frame(): paused")
-        #     return
-
         # Get window dimensions and position in screen coordinates
         pos = imgui.get_window_pos()
         size = imgui.get_window_size()
@@ -242,30 +272,18 @@ class OpenglFrontend(FrontendBase):
         if size.x <= 0 or size.y <= 0:
             return
 
-        self._late_center(size)
-
         # Convert ImGui coordinates to GL coordinates (flip Y)
         gl_y = io.display_size.y - (pos.y + size.y)
         gl.glViewport(int(pos.x), int(gl_y), int(size.x), int(size.y))
         self.handle_events(pos, io)
         self.debug_draw.camera.set_size(size.x, size.y)
-        if self._just_a_single_frame:
-            self.single_step()
-            self._just_a_single_frame = False
-        else:
-            self.update_and_draw(1 / 60)
+
+        # draw the world
+        self.draw_physics()
 
         # Reset viewport
         gl.glViewport(0, 0, int(io.display_size.x), int(io.display_size.y))
 
-        # some potential sleeping here
-        expected_dt = 1 / self.settings.fps
-        dt = time.time() - self._end_of_last_frame_time
-        if dt < expected_dt:
-            sleep_time = expected_dt - dt
-            time.sleep(sleep_time)
-        self._end_of_last_frame_time = time.time()
-
     def create_simulation_window(self):
         weak = weakref.proxy(self)
         window = hello_imgui.DockableWindow()
@@ -376,18 +394,57 @@ class OpenglFrontend(FrontendBase):
         # play/pause toggle-button
         # stop-button
         # single-step button
-        is_paused = self._is_paused
+        is_paused = self.is_paused()
 
         imgui.text("Controls")
         imgui.separator()
-        if ui_button(ICON_PLAY, same_line=False):
-            self._is_paused = not self._is_paused
+
+        play_or_pause_icon = ICON_PLAY if is_paused else ICON_PAUSE
+
+        if ui_button(play_or_pause_icon, same_line=False):
+            if self.is_paused():
+                self.set_running()
+            else:
+                self.set_paused()
         if ui_button(ICON_FORWARD_STEP, enabled=is_paused, same_line=True):
             self._just_a_single_frame = True
         if ui_button(ICON_STOP, same_line=True):
             self.stop()
         imgui.separator()
 
+        # slider for hertz
+        changed, hertz = imgui.slider_float("Hertz", self.settings.hertz, 1.0, 120.0, "%.1f Hz")
+        if changed:
+            self.settings.hertz = hertz
+
+        # slider for amount of substeps
+        changed, substeps = imgui.slider_int(
+            "Substeps", self.settings.substeps, 0, 30, "%d substeps"
+        )
+        if changed:
+            self.settings.substeps = substeps
+
+        max_factor = 4.0
+
+        def mapping(val):
+            if val > 0:
+                return 1.0 + val
+            elif val == 0:
+                return 1.0
+            else:
+                return 1.0 / abs(val - 1)
+
+        changed, val = imgui.slider_float(
+            "Speed",
+            self.speed_ui_val,
+            -max_factor,
+            max_factor,
+            f"{mapping(self.speed_ui_val):.2f}x",
+        )
+        if changed:
+            self.speed_ui_val = val
+            self.settings.speed = mapping(val)
+
     def create_debug_draw_settings_window(self):
         weak = weakref.proxy(self)
         window = hello_imgui.DockableWindow()
@@ -420,17 +477,19 @@ class OpenglFrontend(FrontendBase):
         return docking_params
 
     def on_pre_new_frame(self):
-        if self._center_when_ready:
-            # Center the sample when the debug draw is ready
-            self.center_sample(margin_px=self._margin_px)
-            self._center_when_ready = False
-            self._margin_px = None
-
-        # we can use it to update the sample
-        # todo, meassure time
-        # print("OpenglFrontend.on_pre_new_frame(): updating sample")
-        # self.update_and_draw(1/60)
-        # self.sample.world.step(1/60, 10)
+        self.update_frontend_logic()
+        if self._is_paused and not self._just_a_single_frame:
+            return
+        # update the physics simulation
+        if self.sample is not None:
+            if self._just_a_single_frame:
+                # just a single time step with dt=1 / self.settings.hertz
+                self.update_physics_single_step()
+                self._just_a_single_frame = False
+            else:
+                # this may do as many steps as needed
+                # to catch up with the passed time
+                self.update_physics()
 
     def before_exit(self):
         self.sample.world.destroy()
@@ -443,10 +502,7 @@ class OpenglFrontend(FrontendBase):
         self.debug_draw.draw_joints = True  # self.settings.debug_draw.draw_joints
 
     def center_sample(self, margin_px):
-        if self.debug_draw is None:
-            self._center_when_ready = True
-            self._margin_px = margin_px
-            return
+        assert self.sample is not None, "Sample must be set before centering."
 
         aabb = self.sample.aabb()
 
